import gspread
from google.auth.transport.requests import AuthorizedSession

from gcp_pilot.base import GoogleCloudPilotAPI


class Spreadsheet(GoogleCloudPilotAPI):
    _scopes = [
        'https://www.googleapis.com/auth/spreadsheets',
        'https://www.googleapis.com/auth/drive',
    ]

    def __init__(self, sheet_id: str, **kwargs):
        super().__init__(**kwargs)
        self.sheet_id = sheet_id
        self.spreadsheet = self.client.open_by_key(self.sheet_id)

    def worksheet(self, name: str) -> gspread.Worksheet:
        return self.spreadsheet.worksheet(name)

    @classmethod
    def _client_class(cls, credentials, **kwargs):
        gc = gspread.Client(auth=credentials)
        gc.session = AuthorizedSession(credentials)
        return gc

    @property
    def url(self) -> str:
        return f'https://docs.google.com/spreadsheets/d/{self.sheet_id}/edit'


__all__ = (
    'Spreadsheet',
)



from contextlib import ExitStack
from unittest.mock import patch

from google.oauth2.service_account import Credentials


def patch_auth(project_id: str = 'potato-dev', location: str = 'moon-dark1'):
    # Realistic: actual class to be accepted by clients during validation
    # But fake: with as few attributes as possible, any API call using the credential should fail
    credentials = Credentials(
        service_account_email='chuck@norris.com',
        signer=None,
        token_uri='',
        project_id=project_id,
    )
    managers = [
        patch('google.auth.default', return_value=(credentials, project_id)),
        patch('gcp_pilot.base.GoogleCloudPilotAPI._set_location', return_value=location),
        patch('gcp_pilot.base.AppEngineBasedService._set_location', return_value=location),
    ]
    stack = ExitStack()
    for mgr in managers:
        stack.enter_context(mgr)
    return stack



from gcp_pilot.base import GoogleCloudPilotAPI, DiscoveryMixin, _CACHED_LOCATIONS, AppEngineBasedService


class AppEngine(AppEngineBasedService, DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = [
        'https://www.googleapis.com/auth/appengine.admin',
    ]

    def __init__(self, **kwargs):
        super().__init__(
            serviceName='appengine',
            version='v1',
            cache_discovery=False,
            **kwargs,
        )

    def _set_location(self, location: str = None) -> str:
        return location or self._get_default_location()

    def get_app(self, app_id: str = None):
        app_id = app_id or self.project_id
        return self._execute(
            method=self.client.apps().get,
            appsId=app_id,
        )

    def _get_default_location(self, default_zone: str = '1') -> str:
        location = _CACHED_LOCATIONS.get(self.project_id, None)

        if not location:
            app = self.get_app()
            location = app['locationId']
            try:
                int(location[-1])
            except ValueError:
                location = app['locationId'] + default_zone
            _CACHED_LOCATIONS[self.project_id] = location
        return location


__all__ = (
    'AppEngine',
)



import uuid
from datetime import datetime, timedelta

from google.api_core.exceptions import FailedPrecondition
from google.cloud import tasks_v2
from google.protobuf import timestamp_pb2

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI, AppEngineBasedService


class CloudTasks(AppEngineBasedService, GoogleCloudPilotAPI):
    _client_class = tasks_v2.CloudTasksClient
    DEFAULT_METHOD = tasks_v2.HttpMethod.POST

    def _parent_path(self, project_id: str = None) -> str:
        return f'projects/{project_id or self.project_id}/locations/{self.location}'

    def _queue_path(self, queue: str, project_id: str = None) -> str:
        return self.client.queue_path(
            project=project_id or self.project_id,
            location=self.location,
            queue=queue,
        )

    def _task_path(self, task: str, queue: str, project_id: str = None) -> str:
        return self.client.task_path(
            project=project_id or self.project_id,
            location=self.location,
            queue=queue,
            task=task,
        )

    async def push(
            self,
            task_name: str,
            queue_name: str,
            url: str,
            payload: str,
            method: int = DEFAULT_METHOD,
            delay_in_seconds: int = 0,
            project_id: str = None,
            unique: bool = True,
            use_oidc_auth: bool = True,
    ) -> tasks_v2.Task:
        queue_path = self.client.queue_path(
            project=project_id or self.project_id,
            location=self.location,
            queue=queue_name,
        )
        if unique:
            task_name = f"{task_name}-{str(uuid.uuid4())}"

        task_path = self.client.task_path(
            project=project_id or self.project_id,
            location=self.location,
            queue=queue_name,
            task=task_name,
        )
        task = tasks_v2.Task(
            name=task_path,
            http_request=tasks_v2.HttpRequest(
                http_method=method,
                url=url,
                body=payload.encode(),
                **(self.oidc_token if use_oidc_auth else {}),
            )
        )

        if delay_in_seconds:
            target_date = datetime.utcnow() + timedelta(seconds=delay_in_seconds)
            timestamp = timestamp_pb2.Timestamp()
            timestamp.FromDatetime(target_date)

            task.schedule_time = timestamp

        try:
            response = self.client.create_task(parent=queue_path, task=task)
        except FailedPrecondition as e:
            if 'a queue with this name existed recently' in e.message:
                raise exceptions.DeletedRecently(resource=f"Queue {queue_name}") from e
            if e.message != 'Queue does not exist.':
                raise

            self._create_queue(queue_name=queue_name, project_id=project_id)
            response = self.client.create_task(parent=queue_path, task=task)
        return response

    def _create_queue(
            self,
            queue_name: str,
            project_id: str = None,
    ) -> tasks_v2.Queue:
        parent = self._parent_path(project_id=project_id)
        queue_path = self._queue_path(queue=queue_name, project_id=project_id)

        queue = tasks_v2.Queue(
            name=queue_path,
        )
        return self.client.create_queue(
            parent=parent,
            queue=queue,
        )


__all__ = (
    'CloudTasks',
)
